import malmoenv
import argparse
from pathlib import Path
import time
from PIL import Image
import torch.nn as nn
import torch.optim
import random

# Don't change!
PORT = 9000
SERVER = "127.0.0.1"
ROLE = 0 # Agent role
EXPERIMENTID = "lab8"
RESYNC = 0 # exit and re-sync every N resets (0: off)

# You should not have to change this
MAXRESULTS = 30 # how many results to average over
EPISODE = 0 # Which episode to start with

# Useful parameters for training
EPISODES = 1000 # Number of episodes to run 
MAXSTEPS = 1000 # maximum steps per episode
SAVEIMAGE = 100 # Save an image every N steps

# Which level do you want to run?
DIFFICULTY = 0 # 0 to 4


# Change me!
class QNetwork(nn.Module):
    def __init__(self, D_in, D_out):
        super(QNetwork, self).__init__()
        
        # This is not a good network structure, and just for demonstration purposes!
        self.hidden = nn.Linear(D_in, 3) 
        self.sigmoid = nn.Sigmoid()
        self.out = nn.Linear(3, D_out)
        
    def forward(self, input):
        h = self.hidden(input)
        h = self.sigmoid(h)
        return self.out(h)


def product(seq):
    result = 1
    for s in seq:
        result *= s
    return result

if __name__ == '__main__':
    xml = Path("tasks/findthegoal%d.xml"%DIFFICULTY).read_text()
    env = malmoenv.make()

    env.init(xml, PORT, server=SERVER, role=ROLE, exp_uid=EXPERIMENTID, episode=EPISODE, resync=RESYNC, action_filter = {"move", "turn", "use", "attack", "jump"})
    nactions = len(env.action_space.actions)
    nfeatures = product(env.observation_space.shape)
    Qtarget = QNetwork(nfeatures, nactions)
    Qpolicy = QNetwork(nfeatures, nactions)
    
    # transfer initial weights to target network
    Qtarget.load_state_dict(Qpolicy.state_dict())
    Qtarget.eval()
    
    opt = torch.optim.Adam(Qpolicy.parameters(), lr=0.01)
    minibatch = []
    
    # A variant of the MSELoss that is less sensitive to outliers
    loss_fn = nn.SmoothL1Loss()
    
    eps = 0.1
    GAMMA = 0.7

    BATCHSIZE = 25
    C = 100
    
    iters = 0
    
    first = True
    
    results = []
    
    for i in range(EPISODES):
        #print("reset " + str(i))
        obs = env.reset()
        obs = torch.from_numpy(obs).float().unsqueeze(0)
        steps = 0
        done = False
        while not done and (steps < MAXSTEPS):
        
            # Epsilon-greedy action selection
            if random.random() < eps:
                action = env.action_space.sample()
            else:
                val = Qpolicy(obs)
                (_,action) = val.max(1)
                action = int(action)
                

            oldobs = obs
            obs, reward, done, info = env.step(action)
            obs = torch.from_numpy(obs).float().unsqueeze(0)
            
            # The last observation is empty, we set it to None explicitly so we never use it unintentionally
            if done:
                obs = None
                
            # Save Minibatches
            # store (state, action, reward, newstate)
            minibatch.append((oldobs,action,reward,obs)) 
            
            if len(minibatch) >= BATCHSIZE:
                # turn list of tuples (i.e. [(state, action, reward, newstate)]) into tuple of lists (i.e. ([state], [action], [reward], [newstate]))
                minibatch = list(zip(*minibatch))
                (states,actions,rewards,newstates) = minibatch
                
                # Calculate Q-values for each action for each state
                current_values = Qpolicy(torch.cat(states))
                
                # Select only the actions we actually performed
                action_values = current_values[0,torch.tensor(actions)]
                
                # For the steps where we reach a final state, the "next state" is None 
                # Here we construct two tensors: 
                #                    One tells us which indices are None
                #                    The other gives us the "next states" which are not None
                non_final_mask = torch.tensor(tuple(map(lambda s: s is not None, newstates)), dtype=torch.bool)
                non_final_next_states = torch.cat([s for s in newstates if s is not None])
                
                # Calculate target value using target network: reward + gamma * max Q(s',a')
                target_values = torch.zeros(BATCHSIZE)
                
                # We only have Q values for the next states that exist
                target_values[non_final_mask] = Qtarget(non_final_next_states).max(1)[0].detach()
                target_value = torch.tensor(rewards) + GAMMA*target_values
                
                # We want our network to produce the target value
                loss = loss_fn(action_values, target_value)
                opt.zero_grad()
                loss.backward()
                opt.step()
                
                
                minibatch = []              

            iters += 1
            if iters > C:    
                # transfer weights to target network
                Qtarget.load_state_dict(Qpolicy.state_dict())
                Qtarget.eval()
                iters = 0
            steps += 1
            #print("reward: " + str(reward))
            #print("obs: " + str(obs), len(obs))
            if steps % SAVEIMAGE == 0:
                h, w, d = env.observation_space.shape
                img = Image.fromarray(obs.byte().numpy().reshape(h, w, d))
                img.save('images/image' + str(ROLE) + '_' + str(steps) + '.png')

            time.sleep(.05)
            
        if done:
            results.append(steps)
        if results and len(results)%10 == 0:
            print("avg: %.2f"%(sum(results)*1.0/len(results)))
            results = results[-MAXRESULTS:]

    env.close()
